/*	$Id: xchar.c,v 1.8 2010/01/16 00:31:36 nis Exp $	*/

/*-
 * Copyright (c) 2006 Shingo Nishioka.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if ! defined lint
static char rcsid[] = "$Id: xchar.c,v 1.8 2010/01/16 00:31:36 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>

#include "util.h"

#include <gifnoc.h>

int
xcharp(u)
unsigned int u;
{
	return 
		0x1 <= u && u <= 0xD7FF ||
		0xE000 <= u && u <= 0xFFFD ||
		0x10000 <= u && u <= 0x10FFFF;
}

int
xletterp(u)
unsigned int u;
{
	return 
		0x0041 <= u && u <= 0x005A ||
		0x0061 <= u && u <= 0x007A ||
		0x00C0 <= u && u <= 0x00D6 ||
		0x00D8 <= u && u <= 0x00F6 ||
		0x00F8 <= u && u <= 0x00FF ||
		0x0100 <= u && u <= 0x0131 ||
		0x0134 <= u && u <= 0x013E ||
		0x0141 <= u && u <= 0x0148 ||
		0x014A <= u && u <= 0x017E ||
		0x0180 <= u && u <= 0x01C3 ||
		0x01CD <= u && u <= 0x01F0 ||
		0x01F4 <= u && u <= 0x01F5 ||
		0x01FA <= u && u <= 0x0217 ||
		0x0250 <= u && u <= 0x02A8 ||
		0x02BB <= u && u <= 0x02C1 ||
		u == 0x0386 ||
		0x0388 <= u && u <= 0x038A ||
		u == 0x038C ||
		0x038E <= u && u <= 0x03A1 ||
		0x03A3 <= u && u <= 0x03CE ||
		0x03D0 <= u && u <= 0x03D6 ||
		u == 0x03DA ||
		u == 0x03DC ||
		u == 0x03DE ||
		u == 0x03E0 ||
		0x03E2 <= u && u <= 0x03F3 ||
		0x0401 <= u && u <= 0x040C ||
		0x040E <= u && u <= 0x044F ||
		0x0451 <= u && u <= 0x045C ||
		0x045E <= u && u <= 0x0481 ||
		0x0490 <= u && u <= 0x04C4 ||
		0x04C7 <= u && u <= 0x04C8 ||
		0x04CB <= u && u <= 0x04CC ||
		0x04D0 <= u && u <= 0x04EB ||
		0x04EE <= u && u <= 0x04F5 ||
		0x04F8 <= u && u <= 0x04F9 ||
		0x0531 <= u && u <= 0x0556 ||
		u == 0x0559 ||
		0x0561 <= u && u <= 0x0586 ||
		0x05D0 <= u && u <= 0x05EA ||
		0x05F0 <= u && u <= 0x05F2 ||
		0x0621 <= u && u <= 0x063A ||
		0x0641 <= u && u <= 0x064A ||
		0x0671 <= u && u <= 0x06B7 ||
		0x06BA <= u && u <= 0x06BE ||
		0x06C0 <= u && u <= 0x06CE ||
		0x06D0 <= u && u <= 0x06D3 ||
		u == 0x06D5 ||
		0x06E5 <= u && u <= 0x06E6 ||
		0x0905 <= u && u <= 0x0939 ||
		u == 0x093D ||
		0x0958 <= u && u <= 0x0961 ||
		0x0985 <= u && u <= 0x098C ||
		0x098F <= u && u <= 0x0990 ||
		0x0993 <= u && u <= 0x09A8 ||
		0x09AA <= u && u <= 0x09B0 ||
		u == 0x09B2 ||
		0x09B6 <= u && u <= 0x09B9 ||
		0x09DC <= u && u <= 0x09DD ||
		0x09DF <= u && u <= 0x09E1 ||
		0x09F0 <= u && u <= 0x09F1 ||
		0x0A05 <= u && u <= 0x0A0A ||
		0x0A0F <= u && u <= 0x0A10 ||
		0x0A13 <= u && u <= 0x0A28 ||
		0x0A2A <= u && u <= 0x0A30 ||
		0x0A32 <= u && u <= 0x0A33 ||
		0x0A35 <= u && u <= 0x0A36 ||
		0x0A38 <= u && u <= 0x0A39 ||
		0x0A59 <= u && u <= 0x0A5C ||
		u == 0x0A5E ||
		0x0A72 <= u && u <= 0x0A74 ||
		0x0A85 <= u && u <= 0x0A8B ||
		u == 0x0A8D ||
		0x0A8F <= u && u <= 0x0A91 ||
		0x0A93 <= u && u <= 0x0AA8 ||
		0x0AAA <= u && u <= 0x0AB0 ||
		0x0AB2 <= u && u <= 0x0AB3 ||
		0x0AB5 <= u && u <= 0x0AB9 ||
		u == 0x0ABD ||
		u == 0x0AE0 ||
		0x0B05 <= u && u <= 0x0B0C ||
		0x0B0F <= u && u <= 0x0B10 ||
		0x0B13 <= u && u <= 0x0B28 ||
		0x0B2A <= u && u <= 0x0B30 ||
		0x0B32 <= u && u <= 0x0B33 ||
		0x0B36 <= u && u <= 0x0B39 ||
		u == 0x0B3D ||
		0x0B5C <= u && u <= 0x0B5D ||
		0x0B5F <= u && u <= 0x0B61 ||
		0x0B85 <= u && u <= 0x0B8A ||
		0x0B8E <= u && u <= 0x0B90 ||
		0x0B92 <= u && u <= 0x0B95 ||
		0x0B99 <= u && u <= 0x0B9A ||
		u == 0x0B9C ||
		0x0B9E <= u && u <= 0x0B9F ||
		0x0BA3 <= u && u <= 0x0BA4 ||
		0x0BA8 <= u && u <= 0x0BAA ||
		0x0BAE <= u && u <= 0x0BB5 ||
		0x0BB7 <= u && u <= 0x0BB9 ||
		0x0C05 <= u && u <= 0x0C0C ||
		0x0C0E <= u && u <= 0x0C10 ||
		0x0C12 <= u && u <= 0x0C28 ||
		0x0C2A <= u && u <= 0x0C33 ||
		0x0C35 <= u && u <= 0x0C39 ||
		0x0C60 <= u && u <= 0x0C61 ||
		0x0C85 <= u && u <= 0x0C8C ||
		0x0C8E <= u && u <= 0x0C90 ||
		0x0C92 <= u && u <= 0x0CA8 ||
		0x0CAA <= u && u <= 0x0CB3 ||
		0x0CB5 <= u && u <= 0x0CB9 ||
		u == 0x0CDE ||
		0x0CE0 <= u && u <= 0x0CE1 ||
		0x0D05 <= u && u <= 0x0D0C ||
		0x0D0E <= u && u <= 0x0D10 ||
		0x0D12 <= u && u <= 0x0D28 ||
		0x0D2A <= u && u <= 0x0D39 ||
		0x0D60 <= u && u <= 0x0D61 ||
		0x0E01 <= u && u <= 0x0E2E ||
		u == 0x0E30 ||
		0x0E32 <= u && u <= 0x0E33 ||
		0x0E40 <= u && u <= 0x0E45 ||
		0x0E81 <= u && u <= 0x0E82 ||
		u == 0x0E84 ||
		0x0E87 <= u && u <= 0x0E88 ||
		u == 0x0E8A ||
		u == 0x0E8D ||
		0x0E94 <= u && u <= 0x0E97 ||
		0x0E99 <= u && u <= 0x0E9F ||
		0x0EA1 <= u && u <= 0x0EA3 ||
		u == 0x0EA5 ||
		u == 0x0EA7 ||
		0x0EAA <= u && u <= 0x0EAB ||
		0x0EAD <= u && u <= 0x0EAE ||
		u == 0x0EB0 ||
		0x0EB2 <= u && u <= 0x0EB3 ||
		u == 0x0EBD ||
		0x0EC0 <= u && u <= 0x0EC4 ||
		0x0F40 <= u && u <= 0x0F47 ||
		0x0F49 <= u && u <= 0x0F69 ||
		0x10A0 <= u && u <= 0x10C5 ||
		0x10D0 <= u && u <= 0x10F6 ||
		u == 0x1100 ||
		0x1102 <= u && u <= 0x1103 ||
		0x1105 <= u && u <= 0x1107 ||
		u == 0x1109 ||
		0x110B <= u && u <= 0x110C ||
		0x110E <= u && u <= 0x1112 ||
		u == 0x113C ||
		u == 0x113E ||
		u == 0x1140 ||
		u == 0x114C ||
		u == 0x114E ||
		u == 0x1150 ||
		0x1154 <= u && u <= 0x1155 ||
		u == 0x1159 ||
		0x115F <= u && u <= 0x1161 ||
		u == 0x1163 ||
		u == 0x1165 ||
		u == 0x1167 ||
		u == 0x1169 ||
		0x116D <= u && u <= 0x116E ||
		0x1172 <= u && u <= 0x1173 ||
		u == 0x1175 ||
		u == 0x119E ||
		u == 0x11A8 ||
		u == 0x11AB ||
		0x11AE <= u && u <= 0x11AF ||
		0x11B7 <= u && u <= 0x11B8 ||
		u == 0x11BA ||
		0x11BC <= u && u <= 0x11C2 ||
		u == 0x11EB ||
		u == 0x11F0 ||
		u == 0x11F9 ||
		0x1E00 <= u && u <= 0x1E9B ||
		0x1EA0 <= u && u <= 0x1EF9 ||
		0x1F00 <= u && u <= 0x1F15 ||
		0x1F18 <= u && u <= 0x1F1D ||
		0x1F20 <= u && u <= 0x1F45 ||
		0x1F48 <= u && u <= 0x1F4D ||
		0x1F50 <= u && u <= 0x1F57 ||
		u == 0x1F59 ||
		u == 0x1F5B ||
		u == 0x1F5D ||
		0x1F5F <= u && u <= 0x1F7D ||
		0x1F80 <= u && u <= 0x1FB4 ||
		0x1FB6 <= u && u <= 0x1FBC ||
		u == 0x1FBE ||
		0x1FC2 <= u && u <= 0x1FC4 ||
		0x1FC6 <= u && u <= 0x1FCC ||
		0x1FD0 <= u && u <= 0x1FD3 ||
		0x1FD6 <= u && u <= 0x1FDB ||
		0x1FE0 <= u && u <= 0x1FEC ||
		0x1FF2 <= u && u <= 0x1FF4 ||
		0x1FF6 <= u && u <= 0x1FFC ||
		u == 0x2126 ||
		0x212A <= u && u <= 0x212B ||
		u == 0x212E ||
		0x2180 <= u && u <= 0x2182 ||
		0x3041 <= u && u <= 0x3094 ||
		0x30A1 <= u && u <= 0x30FA ||
		0x3105 <= u && u <= 0x312C ||
		0xAC00 <= u && u <= 0xD7A3 ||
		0x4E00 <= u && u <= 0x9FA5 ||
		u == 0x3007 ||
		0x3021 <= u && u <= 0x3029;
}

int
xdigitp(u)
unsigned int u;
{
	return 
		0x0030 <= u && u <= 0x0039 ||
		0x0660 <= u && u <= 0x0669 ||
		0x06F0 <= u && u <= 0x06F9 ||
		0x0966 <= u && u <= 0x096F ||
		0x09E6 <= u && u <= 0x09EF ||
		0x0A66 <= u && u <= 0x0A6F ||
		0x0AE6 <= u && u <= 0x0AEF ||
		0x0B66 <= u && u <= 0x0B6F ||
		0x0BE7 <= u && u <= 0x0BEF ||
		0x0C66 <= u && u <= 0x0C6F ||
		0x0CE6 <= u && u <= 0x0CEF ||
		0x0D66 <= u && u <= 0x0D6F ||
		0x0E50 <= u && u <= 0x0E59 ||
		0x0ED0 <= u && u <= 0x0ED9 ||
		0x0F20 <= u && u <= 0x0F29;
}

int
xcombiningcharp(u)
unsigned int u;
{
	return 
		0x0300 <= u && u <= 0x0345 ||
		0x0360 <= u && u <= 0x0361 ||
		0x0483 <= u && u <= 0x0486 ||
		0x0591 <= u && u <= 0x05A1 ||
		0x05A3 <= u && u <= 0x05B9 ||
		0x05BB <= u && u <= 0x05BD ||
		u == 0x05BF ||
		0x05C1 <= u && u <= 0x05C2 ||
		u == 0x05C4 ||
		0x064B <= u && u <= 0x0652 ||
		u == 0x0670 ||
		0x06D6 <= u && u <= 0x06DC ||
		0x06DD <= u && u <= 0x06DF ||
		0x06E0 <= u && u <= 0x06E4 ||
		0x06E7 <= u && u <= 0x06E8 ||
		0x06EA <= u && u <= 0x06ED ||
		0x0901 <= u && u <= 0x0903 ||
		u == 0x093C ||
		0x093E <= u && u <= 0x094C ||
		u == 0x094D ||
		0x0951 <= u && u <= 0x0954 ||
		0x0962 <= u && u <= 0x0963 ||
		0x0981 <= u && u <= 0x0983 ||
		u == 0x09BC ||
		u == 0x09BE ||
		u == 0x09BF ||
		0x09C0 <= u && u <= 0x09C4 ||
		0x09C7 <= u && u <= 0x09C8 ||
		0x09CB <= u && u <= 0x09CD ||
		u == 0x09D7 ||
		0x09E2 <= u && u <= 0x09E3 ||
		u == 0x0A02 ||
		u == 0x0A3C ||
		u == 0x0A3E ||
		u == 0x0A3F ||
		0x0A40 <= u && u <= 0x0A42 ||
		0x0A47 <= u && u <= 0x0A48 ||
		0x0A4B <= u && u <= 0x0A4D ||
		0x0A70 <= u && u <= 0x0A71 ||
		0x0A81 <= u && u <= 0x0A83 ||
		u == 0x0ABC ||
		0x0ABE <= u && u <= 0x0AC5 ||
		0x0AC7 <= u && u <= 0x0AC9 ||
		0x0ACB <= u && u <= 0x0ACD ||
		0x0B01 <= u && u <= 0x0B03 ||
		u == 0x0B3C ||
		0x0B3E <= u && u <= 0x0B43 ||
		0x0B47 <= u && u <= 0x0B48 ||
		0x0B4B <= u && u <= 0x0B4D ||
		0x0B56 <= u && u <= 0x0B57 ||
		0x0B82 <= u && u <= 0x0B83 ||
		0x0BBE <= u && u <= 0x0BC2 ||
		0x0BC6 <= u && u <= 0x0BC8 ||
		0x0BCA <= u && u <= 0x0BCD ||
		u == 0x0BD7 ||
		0x0C01 <= u && u <= 0x0C03 ||
		0x0C3E <= u && u <= 0x0C44 ||
		0x0C46 <= u && u <= 0x0C48 ||
		0x0C4A <= u && u <= 0x0C4D ||
		0x0C55 <= u && u <= 0x0C56 ||
		0x0C82 <= u && u <= 0x0C83 ||
		0x0CBE <= u && u <= 0x0CC4 ||
		0x0CC6 <= u && u <= 0x0CC8 ||
		0x0CCA <= u && u <= 0x0CCD ||
		0x0CD5 <= u && u <= 0x0CD6 ||
		0x0D02 <= u && u <= 0x0D03 ||
		0x0D3E <= u && u <= 0x0D43 ||
		0x0D46 <= u && u <= 0x0D48 ||
		0x0D4A <= u && u <= 0x0D4D ||
		u == 0x0D57 ||
		u == 0x0E31 ||
		0x0E34 <= u && u <= 0x0E3A ||
		0x0E47 <= u && u <= 0x0E4E ||
		u == 0x0EB1 ||
		0x0EB4 <= u && u <= 0x0EB9 ||
		0x0EBB <= u && u <= 0x0EBC ||
		0x0EC8 <= u && u <= 0x0ECD ||
		0x0F18 <= u && u <= 0x0F19 ||
		u == 0x0F35 ||
		u == 0x0F37 ||
		u == 0x0F39 ||
		u == 0x0F3E ||
		u == 0x0F3F ||
		0x0F71 <= u && u <= 0x0F84 ||
		0x0F86 <= u && u <= 0x0F8B ||
		0x0F90 <= u && u <= 0x0F95 ||
		u == 0x0F97 ||
		0x0F99 <= u && u <= 0x0FAD ||
		0x0FB1 <= u && u <= 0x0FB7 ||
		u == 0x0FB9 ||
		0x20D0 <= u && u <= 0x20DC ||
		u == 0x20E1 ||
		0x302A <= u && u <= 0x302F ||
		u == 0x3099 ||
		u == 0x309A;
}

int
xextenderp(u)
unsigned int u;
{
	return 
		u == 0x00B7 ||
		u == 0x02D0 ||
		u == 0x02D1 ||
		u == 0x0387 ||
		u == 0x0640 ||
		u == 0x0E46 ||
		u == 0x0EC6 ||
		u == 0x3005 ||
		0x3031 <= u && u <= 0x3035 ||
		0x309D <= u && u <= 0x309E ||
		0x30FC <= u && u <= 0x30FE;
}
